@@ -168,12 +168,13 @@ public final class JAXBUtils {
     }
     
     /** 
-     * Checks if the specified word is a Java keyword (as of 1.5).
+     * Checks if the specified word is a Java keyword (as defined in JavaUtils).
      * 
      * @param word the word to check.
      * @return true if the word is a keyword.
+     * @see org.apache.cxf.helpers.JavaUtils
      */
-    public static boolean isJavaKeyword(String word) {
+    protected static boolean isJavaKeyword(String word) {
         return JavaUtils.isJavaKeyword(word);
     }
 
@@ -194,7 +195,7 @@ public final class JAXBUtils {
     
     /**
      * Generates a Java package name from a URI according to the
-     * algorithm outlined in JAXB 2.0.
+     * algorithm outlined in Appendix D of JAXB (2.0+).
      * 
      * @param namespaceURI the namespace URI.
      * @return the package name.
@@ -203,13 +204,15 @@ public final class JAXBUtils {
        
         StringBuilder packageName = new StringBuilder();
         String authority = uri.getAuthority();
-        if (authority == null && "urn".equals(uri.getScheme())) {
+        String scheme = uri.getScheme();
+        if (authority == null && "urn".equals(scheme)) {
             authority = uri.getSchemeSpecificPart();
         }
         
         if (null != authority && !"".equals(authority)) {
-            if ("urn".equals(uri.getScheme())) {
+            if ("urn".equals(scheme)) {
                 packageName.append(authority);
+                /* JAXB 2.2 D.5.1, Rule #5 */
                 for (int i = 0; i < packageName.length(); i++) {
                     if (packageName.charAt(i) == '-') {
                         packageName.setCharAt(i, '.');
@@ -247,17 +250,29 @@ public final class JAXBUtils {
                     }
                     packageName.insert(0, normalizePackageNamePart(token));
                 }
-               
             }
+            
+            if (!("http".equalsIgnoreCase(scheme) || "urn".equalsIgnoreCase(scheme))) {
+                packageName.insert(0, ".");
+                packageName.insert(0, normalizePackageNamePart(scheme));
+            }
+            
         }
 
         String path = uri.getPath();
         if (path == null) {
             path = "";
         }
+        /* JAXB 2.2 D.5.1 Rule 2 - remove trailing .??, .???, or .html only. */
         int index = path.lastIndexOf('.');
         if (index < 0) {
             index = path.length();
+        } else {
+            String ending = path.substring(index + 1);
+            if (ending.length() < 2 || (ending.length() > 3 
+                && !"html".equalsIgnoreCase(ending))) {
+                index = path.length();
+            }
         }
         StringTokenizer st = new StringTokenizer(path.substring(0, index), "/");
         while (st.hasMoreTokens()) {
@@ -297,7 +312,7 @@ public final class JAXBUtils {
     
     /**
      * Converts an XML name to a Java identifier according to the mapping
-     * algorithm outlines in the JAXB specification
+     * algorithm outlined in the JAXB specification
      * 
      * @param name the XML name
      * @return the Java identifier
@@ -664,6 +679,45 @@ public final class JAXBUtils {
             throw new JAXBException(ex);
         }
     }
+
+    public static SchemaCompiler createSchemaCompilerWithDefaultAllocator(Set<String> allocatorSet) {
+        
+        try {
+            SchemaCompiler compiler = JAXBUtils.createSchemaCompiler();
+            Object allocator = ReflectionInvokationHandler
+                .createProxyWrapper(new DefaultClassNameAllocator(allocatorSet),
+                                JAXBUtils.getParamClass(compiler, "setClassNameAllocator"));
+
+            compiler.setClassNameAllocator(allocator);
+            return compiler;    
+        } catch (JAXBException e1) {
+            throw new IllegalStateException("Unable to create schema compiler", e1);
+        }
+        
+    }
+    
+    public static void logGeneratedClassNames(Logger logger, JCodeModel codeModel) {
+        if (!logger.isLoggable(Level.INFO)) {
+            return;
+        }
+        
+        StringBuilder sb = new StringBuilder();
+        boolean first = true;
+        for (Iterator<JPackage> itr = codeModel.packages(); itr.hasNext();) {
+            JPackage package1 = itr.next();
+            
+            for (Iterator<JDefinedClass> citr = package1.classes(); citr.hasNext();) {
+                if (!first) {
+                    sb.append(", ");
+                } else {
+                    first = false;
+                }
+                sb.append(citr.next().fullName());
+            }
+        }
+        
+        logger.log(Level.INFO, "Created classes: " + sb.toString());
+    }
     
     public static Object createFileCodeWriter(File f) throws JAXBException {
         try {
@@ -724,7 +778,7 @@ public final class JAXBUtils {
     }
     
     public static String getPackageNamespace(Class<?> cls) {
-        Package p = Package.getPackage(PackageUtils.getPackageName(cls));
+        Package p = cls.getPackage();
         if (p != null) {
             javax.xml.bind.annotation.XmlSchema schemaAnn = 
                 p.getAnnotation(javax.xml.bind.annotation.XmlSchema.class);
@@ -735,8 +789,17 @@ public final class JAXBUtils {
         return null;
     }
     
-    public static void scanPackages(Set<Class<?>> classes, 
+    public static void scanPackages(Set<Class<?>> classes,
+                                    Class[] extraClass,
                                     Map<Package, CachedClass> objectFactoryCache) {
+        
+        // add user extra class into jaxb context
+        if (extraClass != null && extraClass.length > 0) {
+            for (Class clz : extraClass) {
+                classes.add(clz);
+            }
+        }
+        
         // try and read any jaxb.index files that are with the other classes.
         // This should
         // allow loading of extra classes (such as subclasses for inheritance
@@ -779,9 +842,9 @@ public final class JAXBUtils {
         }
         for (Map.Entry<String, InputStream> entry : packages.entrySet()) {
             if (entry.getValue() != null) {
+                BufferedReader reader = null;
                 try {
-                    BufferedReader reader = new BufferedReader(new InputStreamReader(entry.getValue(),
-                                                                                     "UTF-8"));
+                    reader = new BufferedReader(new InputStreamReader(entry.getValue(), "UTF-8"));
                     String pkg = entry.getKey();
                     ClassLoader loader = packageLoaders.get(pkg);
                     if (!StringUtils.isEmpty(pkg)) {
@@ -808,7 +871,10 @@ public final class JAXBUtils {
                     // ignore
                 } finally {
                     try {
-                        entry.getValue().close();
+                        if (reader != null) {
+                            reader.close();
+                            entry.getValue().close();
+                        }
                     } catch (Exception e) {
                         // ignore
                     }
@@ -831,6 +897,32 @@ public final class JAXBUtils {
         }
     }
     
+    public static class DefaultClassNameAllocator {
+        private final Set<String> typesClassNames;
+
+        public DefaultClassNameAllocator() {
+            this(new HashSet<String>());
+        }
+        
+        public DefaultClassNameAllocator(Set<String> set) {
+            typesClassNames = set;
+        }
+
+        public String assignClassName(String packageName, String className) {
+            String fullClassName = className;
+            String fullPckClass = packageName + "." + fullClassName;
+            int cnt = 0;
+            while (typesClassNames.contains(fullPckClass)) {
+                cnt++;
+                fullClassName = className + cnt;
+                fullPckClass = packageName + "." + fullClassName;
+            }
+            typesClassNames.add(fullPckClass);
+            return fullClassName;
+        }
+       
+    }
+    
     public static interface SchemaCompiler {
 
         void setErrorListener(Object elForRun);
