@@ -26,6 +26,7 @@ import java.io.StringReader;
 import java.io.StringWriter;
 import java.lang.annotation.Annotation;
 import java.lang.reflect.Method;
+import java.lang.reflect.Type;
 import java.net.URI;
 import java.util.ArrayList;
 import java.util.Arrays;
@@ -109,11 +110,13 @@ public class WadlGenerator implements RequestHandler {
     private static final String JAXB_DEFAULT_NAMESPACE = "##default";
     private static final String JAXB_DEFAULT_NAME = "##default";
     private static final String CLASSPATH_PREFIX = "classpath:";
-    
+    private static final String DEFAULT_NS_PREFIX = "prefix";
+
     private String wadlNamespace;
     private boolean ignoreMessageWriters = true;
     private boolean singleResourceMultipleMethods = true;
     private boolean useSingleSlashResource;
+    private boolean ignoreForwardSlash;
     private boolean addResourceAndMethodIds;
     
     private boolean useJaxbContextForQnames = true;
@@ -123,7 +126,9 @@ public class WadlGenerator implements RequestHandler {
     private Map<String, List<String>> externalQnamesMap; 
     private ElementQNameResolver resolver;
     private List<String> privateAddresses;
-    
+    private String applicationTitle;
+    private String nsPrefix = DEFAULT_NS_PREFIX;
+
     public WadlGenerator() {
         
     }
@@ -191,6 +196,7 @@ public class WadlGenerator implements RequestHandler {
         
         sbGrammars.append("</grammars>");
         sbMain.append(">");
+        handleApplicationDocs(sbMain);
         sbMain.append(sbGrammars.toString());
         sbMain.append(sbResources.toString());
         sbMain.append("</application>");
@@ -235,7 +241,7 @@ public class WadlGenerator implements RequestHandler {
             OperationResourceInfo ori = sortedOps.get(i); 
             
             if (ori.getHttpMethod() == null) {
-                Class<?> cls = ori.getMethodToInvoke().getReturnType();
+                Class<?> cls = getMethod(ori).getReturnType();
                 ClassResourceInfo subcri = cri.findResource(cls, cls);
                 if (subcri != null && !visitedResources.contains(subcri)) {
                     startResourceTag(sb, subcri.getServiceClass(), ori.getURITemplate().getValue());
@@ -256,7 +262,7 @@ public class WadlGenerator implements RequestHandler {
     }
     
     private void startResourceTag(StringBuilder sb, Class<?> serviceClass, String path) {
-        sb.append("<resource path=\"").append(path).append("\"");
+        sb.append("<resource path=\"").append(getPath(path)).append("\"");
         if (addResourceAndMethodIds) {
             QName jaxbQname = null;
             if (useJaxbContextForQnames) {
@@ -271,10 +277,18 @@ public class WadlGenerator implements RequestHandler {
         sb.append(">");
     }
     
+    private String getPath(String path) {
+        if (ignoreForwardSlash && path.startsWith("/") && path.length() > 0) {
+            return path.substring(1);
+        } else {
+            return path;
+        }
+    }
+    
     private void startMethodTag(StringBuilder sb, OperationResourceInfo ori) {
         sb.append("<method name=\"").append(ori.getHttpMethod()).append("\"");
         if (addResourceAndMethodIds) {
-            sb.append(" id=\"").append(ori.getMethodToInvoke().getName()).append("\"");
+            sb.append(" id=\"").append(getMethod(ori).getName()).append("\"");
         }
         sb.append(">");
     }
@@ -300,18 +314,18 @@ public class WadlGenerator implements RequestHandler {
                     path = path.substring(1); 
                 }
             }
-            sb.append("<resource path=\"").append(path).append("\">");
+            sb.append("<resource path=\"").append(getPath(path)).append("\">");
             handlePathAndMatrixParams(sb, ori);
         } else if (index == 0) {
             handlePathAndMatrixParams(sb, ori);
         }
         
         startMethodTag(sb, ori);
-        handleDocs(ori.getAnnotatedMethod().getAnnotations(), sb);
-        if (ori.getMethodToInvoke().getParameterTypes().length != 0) {
+        handleDocs(getMethod(ori).getAnnotations(), sb);
+        if (getMethod(ori).getParameterTypes().length != 0) {
             sb.append("<request>");
             if (isFormRequest(ori)) {
-                handleRepresentation(sb, jaxbTypes, qnameResolver, clsMap, ori, null, false);
+                handleFormRepresentation(sb, jaxbTypes, qnameResolver, clsMap, ori, getFormClass(ori));
             } else {
                 for (Parameter p : ori.getParameters()) {        
                     handleParameter(sb, jaxbTypes, qnameResolver, clsMap, ori, p);             
@@ -320,15 +334,15 @@ public class WadlGenerator implements RequestHandler {
             sb.append("</request>");
         }
         sb.append("<response");
-        boolean isVoid = void.class == ori.getMethodToInvoke().getReturnType();
+        boolean isVoid = void.class == getMethod(ori).getReturnType();
         if (isVoid) {
-            boolean oneway = ori.getMethodToInvoke().getAnnotation(Oneway.class) != null;
+            boolean oneway = getMethod(ori).getAnnotation(Oneway.class) != null;
             sb.append(" status=\"" + (oneway ? 202 : 204) + "\"");
         }
         sb.append(">");
-        if (void.class != ori.getMethodToInvoke().getReturnType()) {
+        if (void.class != getMethod(ori).getReturnType()) {
             handleRepresentation(sb, jaxbTypes, qnameResolver, clsMap, ori,
-                                 ori.getMethodToInvoke().getReturnType(), false);
+                                 getMethod(ori).getReturnType(), false);
         }
         sb.append("</response>");
         
@@ -400,7 +414,7 @@ public class WadlGenerator implements RequestHandler {
     private void handleParameter(StringBuilder sb, Set<Class<?>> jaxbTypes, 
                                  ElementQNameResolver qnameResolver, 
                                  Map<Class<?>, QName> clsMap, OperationResourceInfo ori, Parameter pm) {
-        Class<?> cls = ori.getMethodToInvoke().getParameterTypes()[pm.getIndex()];
+        Class<?> cls = getMethod(ori).getParameterTypes()[pm.getIndex()];
         if (pm.getType() == ParameterType.REQUEST_BODY) {
             handleRepresentation(sb, jaxbTypes, qnameResolver, clsMap, ori, cls, true);
             return;
@@ -426,19 +440,20 @@ public class WadlGenerator implements RequestHandler {
         if (inbound) {
             for (Parameter pm : ori.getParameters()) {
                 if (pm.getType() == ParameterType.REQUEST_BODY) {
-                    return ori.getAnnotatedMethod().getParameterAnnotations()[pm.getIndex()];
+                    return getMethod(ori).getParameterAnnotations()[pm.getIndex()];
                 }
             }
             return new Annotation[]{};
         } else {
-            return ori.getAnnotatedMethod().getDeclaredAnnotations();
+            return getMethod(ori).getDeclaredAnnotations();
         }
     }
     
     private void writeParam(StringBuilder sb, Parameter pm, OperationResourceInfo ori) {
-        Class<?> type = ori.getMethodToInvoke().getParameterTypes()[pm.getIndex()];
+        Class<?> type = getMethod(ori).getParameterTypes()[pm.getIndex()];
         if (!"".equals(pm.getName())) {
-            doWriteParam(sb, pm, type, pm.getName());
+            doWriteParam(sb, pm, type, getMethod(ori).getGenericParameterTypes()[pm.getIndex()],
+                         pm.getName());
         } else {
             doWriteBeanParam(sb, type, pm, null);
         }
@@ -451,26 +466,31 @@ public class WadlGenerator implements RequestHandler {
             if (parentName != null) {
                 name = parentName + "." + name;
             }
-            if (InjectionUtils.isPrimitive(entry.getValue())) {
-                doWriteParam(sb, entry.getKey(), entry.getValue(), name);
+            Class<?> paramCls = entry.getValue();
+            boolean isPrimitive = InjectionUtils.isPrimitive(paramCls);
+            if (isPrimitive || InjectionUtils.isSupportedCollectionOrArray(paramCls)) {
+                doWriteParam(sb, entry.getKey(), paramCls, paramCls, name);
             } else {
-                doWriteBeanParam(sb, entry.getValue(), entry.getKey(), name);
+                doWriteBeanParam(sb, paramCls, entry.getKey(), name);
             }
         }
     }
-    
-    protected void doWriteParam(StringBuilder sb, Parameter pm, Class<?> type, String paramName) {
-        if (ParameterType.REQUEST_BODY == pm.getType()) {
-            return;
-        }
+
+    protected void doWriteParam(StringBuilder sb, Parameter pm, Class<?> type, 
+                                Type genericType, String paramName) {
         sb.append("<param name=\"").append(paramName).append("\" ");
         String style = ParameterType.PATH == pm.getType() ? "template" 
                        : ParameterType.FORM == pm.getType() ? "query"
+                       : ParameterType.REQUEST_BODY == pm.getType() ? "plain"    
                        : pm.getType().toString().toLowerCase();
         sb.append("style=\"").append(style).append("\"");
         if (pm.getDefaultValue() != null) {
             sb.append(" default=\"").append(pm.getDefaultValue()).append("\"");
         }
+        if (InjectionUtils.isSupportedCollectionOrArray(type)) {
+            type = InjectionUtils.getActualType(genericType);
+            sb.append(" repeating=\"true\"");
+        }
         String value = XmlSchemaPrimitiveUtils.getSchemaRepresentation(type);
         if (value != null) {
             sb.append(" type=\"").append(value).append("\"");
@@ -483,30 +503,61 @@ public class WadlGenerator implements RequestHandler {
                                       Map<Class<?>, QName> clsMap, OperationResourceInfo ori, 
                                       Class<?> type, boolean inbound) {
         List<MediaType> types = inbound ? ori.getConsumeTypes() : ori.getProduceTypes();
-        if (types.size() == 1 && types.get(0).equals(MediaType.WILDCARD_TYPE)
-            && (type == null || MultivaluedMap.class.isAssignableFrom(type))) {
+        if (MultivaluedMap.class.isAssignableFrom(type)) {
             types = Collections.singletonList(MediaType.APPLICATION_FORM_URLENCODED_TYPE);
-        }
-        if (type != null) {
-            for (MediaType mt : types) {
-                if (InjectionUtils.isPrimitive(type)) {
-                    String rep = XmlSchemaPrimitiveUtils.getSchemaRepresentation(type);
-                    String value = rep == null ? type.getSimpleName() : rep;
-                    sb.append("<!-- Primitive type : " + value + " -->");
-                }
-                sb.append("<representation");
-                sb.append(" mediaType=\"").append(mt.toString()).append("\"");
-                
-                type = getActualJaxbType(type, ori.getAnnotatedMethod(), inbound);
+        } else if (isWildcard(types)) {
+            types = Collections.singletonList(MediaType.APPLICATION_OCTET_STREAM_TYPE);
+        } 
+        
+        boolean isPrimitive = InjectionUtils.isPrimitive(type);
+        for (MediaType mt : types) {
+            
+            sb.append("<representation");
+            sb.append(" mediaType=\"").append(mt.toString()).append("\"");
+
+            if (isPrimitive) {
+                sb.append(">");    
+                Parameter p = inbound ? getRequestBodyParam(ori) 
+                    : new Parameter(ParameterType.REQUEST_BODY, 0, "result"); 
+                doWriteParam(sb, p, type, type, p.getName() == null ? "request" : p.getName());
+                sb.append("</representation>");
+            } else  { 
+                type = getActualJaxbType(type, getMethod(ori), inbound);
                 if (qnameResolver != null && mt.getSubtype().contains("xml") && jaxbTypes.contains(type)) {
                     generateQName(sb, qnameResolver, clsMap, type,
                                   getBodyAnnotations(ori, inbound));
                 }
                 sb.append("/>");
             }
+        }
+        
+    }
+    
+    private Parameter getRequestBodyParam(OperationResourceInfo ori) {
+        for (Parameter p : ori.getParameters()) {
+            if (p.getType() == ParameterType.REQUEST_BODY) {
+                return p;
+            }
+        }
+        throw new IllegalStateException();
+    }
+    
+    private boolean isWildcard(List<MediaType> types) {
+        return types.size() == 1 && types.get(0).equals(MediaType.WILDCARD_TYPE);
+    }
+    
+    private void handleFormRepresentation(StringBuilder sb, Set<Class<?>> jaxbTypes,
+                                      ElementQNameResolver qnameResolver,
+                                      Map<Class<?>, QName> clsMap, OperationResourceInfo ori,
+                                      Class<?> type) {
+        if (type != null) {
+            handleRepresentation(sb, jaxbTypes, qnameResolver, clsMap, ori, type, true);
         } else { 
+            List<MediaType> types = ori.getConsumeTypes();
+            MediaType formType = isWildcard(types) ? MediaType.APPLICATION_FORM_URLENCODED_TYPE 
+                : types.get(0); 
             sb.append("<representation");
-            sb.append(" mediaType=\"").append(types.get(0).toString()).append("\">");
+            sb.append(" mediaType=\"").append(formType).append("\">");
             for (Parameter pm : ori.getParameters()) {
                 writeParam(sb, pm, ori);
             }
@@ -644,7 +695,7 @@ public class WadlGenerator implements RequestHandler {
         }    
         if (prefix == null) {
             int size = new HashSet<QName>(clsMap.values()).size();
-            prefix = "prefix" + (size + 1);
+            prefix = nsPrefix + (size + 1);
         }
         return prefix;
     }
@@ -652,13 +703,23 @@ public class WadlGenerator implements RequestHandler {
 
     private boolean isFormRequest(OperationResourceInfo ori) {
         for (Parameter p : ori.getParameters()) {
-            if (p.getType() == ParameterType.FORM) {
+            if (p.getType() == ParameterType.FORM
+                || p.getType() == ParameterType.REQUEST_BODY 
+                && getMethod(ori).getParameterTypes()[p.getIndex()] == MultivaluedMap.class) {
                 return true;
             }
         }
         return false;
     }
     
+    private Class<?> getFormClass(OperationResourceInfo ori) {
+        if (ori.getParameters().get(0).getType() == ParameterType.FORM) {
+            return null;
+        } else {
+            return MultivaluedMap.class;
+        }
+    }
+
     // TODO : can we reuse this block with JAXBBinding somehow ?
     public boolean addSchemaDocument(SchemaCollection col, 
                                      Document d,
@@ -757,6 +818,12 @@ public class WadlGenerator implements RequestHandler {
         this.ignoreMessageWriters = ignoreMessageWriters;
     }
 
+    private void handleApplicationDocs(StringBuilder sbApp) {
+        if (applicationTitle != null) {
+            sbApp.append("<doc title=\"" + applicationTitle + "\"/>");
+        }
+    }
+    
     private void handleDocs(Annotation[] anns, StringBuilder sb) {
         for (Annotation a : anns) {
             if (a.annotationType() == Description.class) {
@@ -1098,6 +1165,23 @@ public class WadlGenerator implements RequestHandler {
         this.addResourceAndMethodIds = addResourceAndMethodIds;
     }
 
+    private Method getMethod(OperationResourceInfo ori) {
+        Method annMethod = ori.getAnnotatedMethod();
+        return annMethod != null ? annMethod : ori.getMethodToInvoke();
+    }
+    
+    public void setApplicationTitle(String applicationTitle) {
+        this.applicationTitle = applicationTitle;
+    }
+
+    public void setNamespacePrefix(String prefix) {
+        this.nsPrefix = prefix;
+    }
+
+    public void setIgnoreForwardSlash(boolean ignoreForwardSlash) {
+        this.ignoreForwardSlash = ignoreForwardSlash;
+    }
+
     private static class SchemaConverter extends DelegatingXMLStreamWriter {
         private static final String SCHEMA_LOCATION = "schemaLocation";
         private Map<String, String> locsMap;    
