@@ -302,9 +302,14 @@ public final class JAXRSUtils {
             new TreeMap<OperationResourceInfo, MultivaluedMap<String, String>>(
                 new OperationResourceInfoComparator(message, httpMethod));
 
-        MediaType requestType = requestContentType == null 
+        MediaType requestType;
+        try {
+            requestType = requestContentType == null
                                 ? ALL_TYPES : MediaType.valueOf(requestContentType);
-        
+        } catch (IllegalArgumentException ex) {
+            throw new WebApplicationException(ex, 415);
+        }
+
         int pathMatched = 0;
         int methodMatched = 0;
         int consumeMatched = 0;
@@ -392,6 +397,7 @@ public final class JAXRSUtils {
         org.apache.cxf.common.i18n.Message errorMsg = 
             new org.apache.cxf.common.i18n.Message(name, 
                                                    BUNDLE,
+                                                   message.get(Message.REQUEST_URI),
                                                    path,
                                                    httpMethod,
                                                    requestType.toString(),
@@ -399,11 +405,16 @@ public final class JAXRSUtils {
         if (!"OPTIONS".equalsIgnoreCase(httpMethod) && logNow) {
             LOG.warning(errorMsg.toString());
         }
-        ResponseBuilder rb = createResponseBuilder(resource, status, methodMatched == 0);
-        throw new WebApplicationException(rb.build());
+        Response response = createResponse(resource, status, methodMatched == 0);
+        throw new WebApplicationException(response);
         
     }    
     
+    public static boolean noResourceMethodForOptions(Response exResponse, String httpMethod) {
+        return exResponse != null && exResponse.getStatus() == 405 
+            && "OPTIONS".equalsIgnoreCase(httpMethod);
+    }
+    
     private static void logNoMatchMessage(OperationResourceInfo ori, 
         String path, String httpMethod, MediaType requestType, List<MediaType> acceptContentTypes) {
         if (!LOG.isLoggable(Level.FINE)) {
@@ -424,7 +435,7 @@ public final class JAXRSUtils {
         LOG.fine(errorMsg.toString());
     }
 
-    public static ResponseBuilder createResponseBuilder(ClassResourceInfo cri, int status, boolean addAllow) {
+    public static Response createResponse(ClassResourceInfo cri, int status, boolean addAllow) {
         ResponseBuilder rb = Response.status(status);
         if (addAllow) {
             Set<String> allowedMethods = cri.getAllowedMethods();
@@ -439,7 +450,7 @@ public final class JAXRSUtils {
                 rb.header("Allow", "HEAD");
             }
         }
-        return rb;
+        return rb.build();
     }
     
     private static boolean matchHttpMethod(String expectedMethod, String httpMethod) {
