@@ -19,13 +19,14 @@
 
 package org.apache.cxf.ws.addressing;
 
+import java.io.IOException;
+import java.io.InputStream;
+import java.net.HttpURLConnection;
 import java.text.MessageFormat;
 import java.util.Collection;
 import java.util.Iterator;
 import java.util.List;
-import java.util.Map;
 import java.util.ResourceBundle;
-import java.util.concurrent.ConcurrentHashMap;
 import java.util.logging.Level;
 import java.util.logging.Logger;
 
@@ -33,17 +34,25 @@ import javax.wsdl.extensions.ExtensibilityElement;
 import javax.xml.namespace.QName;
 import javax.xml.ws.WebFault;
 
+import org.apache.cxf.Bus;
 import org.apache.cxf.binding.soap.SoapBindingConstants;
 import org.apache.cxf.binding.soap.SoapFault;
+import org.apache.cxf.binding.soap.interceptor.SoapActionInInterceptor;
 import org.apache.cxf.binding.soap.model.SoapOperationInfo;
 import org.apache.cxf.common.i18n.BundleUtils;
 import org.apache.cxf.common.logging.LogUtils;
 import org.apache.cxf.common.util.StringUtils;
+import org.apache.cxf.endpoint.Client;
+import org.apache.cxf.endpoint.ClientLifeCycleListener;
+import org.apache.cxf.endpoint.ClientLifeCycleManager;
 import org.apache.cxf.endpoint.Endpoint;
-import org.apache.cxf.helpers.CastUtils;
+import org.apache.cxf.feature.AbstractFeature;
+import org.apache.cxf.helpers.IOUtils;
 import org.apache.cxf.interceptor.Fault;
 import org.apache.cxf.interceptor.OneWayProcessorInterceptor;
+import org.apache.cxf.io.CachedOutputStream;
 import org.apache.cxf.message.Exchange;
+import org.apache.cxf.message.ExchangeImpl;
 import org.apache.cxf.message.FaultMode;
 import org.apache.cxf.message.Message;
 import org.apache.cxf.message.MessageUtils;
@@ -57,10 +66,15 @@ import org.apache.cxf.service.model.OperationInfo;
 import org.apache.cxf.service.model.UnwrappedOperationInfo;
 import org.apache.cxf.transport.Conduit;
 import org.apache.cxf.transport.Destination;
+import org.apache.cxf.transport.DestinationFactory;
+import org.apache.cxf.transport.DestinationFactoryManager;
+import org.apache.cxf.transport.MessageObserver;
+import org.apache.cxf.transport.Observable;
 import org.apache.cxf.ws.addressing.VersionTransformer.Names200408;
 import org.apache.cxf.ws.addressing.policy.MetadataConstants;
 import org.apache.cxf.ws.policy.AssertionInfo;
 import org.apache.cxf.ws.policy.AssertionInfoMap;
+import org.apache.cxf.wsdl.EndpointReferenceUtils;
 
 
 /**
@@ -72,25 +86,40 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
     public static final String ADDRESSING_DISABLED = MAPAggregator.class.getName() + ".addressingDisabled";
     public static final String DECOUPLED_DESTINATION = MAPAggregator.class.getName() 
         + ".decoupledDestination";
-
+    public static final String ACTION_VERIFIED = MAPAggregator.class.getName() + ".actionVerified";
+    
     private static final Logger LOG = 
         LogUtils.getL7dLogger(MAPAggregator.class);
     private static final ResourceBundle BUNDLE = LOG.getResourceBundle();
 
-    
+    private static final ClientLifeCycleListener DECOUPLED_DEST_CLEANER = new ClientLifeCycleListener() {
+        public void clientCreated(Client client) {
+            //ignore
+        }
+
+        public void clientDestroyed(Client client) {
+            Destination dest = client.getEndpoint().getEndpointInfo()
+                .getProperty(DECOUPLED_DESTINATION, Destination.class);
+            if (dest != null) {
+                dest.setMessageObserver(null);
+                dest.shutdown();
+            }
+        }
+        
+    };
 
     /**
-     * REVISIT: map usage implies that the *same* interceptor instance 
-     * is used in all chains.
+     * The cache to use for enforcing uniqueness.  Defaults to {@link DefaultMessageIdCache}.
      */
-    protected final Map<String, String> messageIDs = 
-        new ConcurrentHashMap<String, String>();
+    private MessageIdCache messageIdCache = new DefaultMessageIdCache();
     
     private boolean usingAddressingAdvisory = true;
     private boolean addressingRequired;
 
     private boolean allowDuplicates = true;
     
+    private String addressingResponses = "ALL";
+    
     /**
      * Constructor.
      */
@@ -155,7 +184,39 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
         addressingRequired = required;
     }
     
+    /**
+     * Sets Addresing Response 
+     *
+     */
+    public void setAddressingResponses(String responses) {
+        addressingResponses = responses;
+    }
     
+    /**
+     * Returns the cache used to enforce duplicate message IDs when
+     * {@link #allowDuplicates()} returns {@code false}.
+     *
+     * @return the cache used to enforce duplicate message IDs
+     */
+    public MessageIdCache getMessageIdCache() {
+        return messageIdCache;
+    }
+
+    /**
+     * Sets the cache used to enforce duplicate message IDs when
+     * {@link #allowDuplicates()} returns {@code false}.
+     *
+     * @param messageIdCache the cache to use
+     *
+     * @throws NullPointerException if {@code messageIdCache} is {@code null}
+     */
+    public void setMessageIdCache(MessageIdCache messageIdCache) {
+        if (messageIdCache == null) {
+            throw new NullPointerException("messageIdCache cannot be null.");
+        }
+        this.messageIdCache = messageIdCache;
+    }
+
     /**
      * Invoked for normal processing of inbound and outbound messages.
      *
@@ -300,7 +361,49 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
         } 
         return false;
     }
-    
+
+   
+    private WSAddressingFeature getWSAddressingFeature(Message message) {
+        if (message.getExchange() != null && message.getExchange().getEndpoint() != null) {
+            Endpoint endpoint = message.getExchange().getEndpoint();
+            if (endpoint.getActiveFeatures() != null) {
+                for (AbstractFeature feature : endpoint.getActiveFeatures()) {
+                    if (feature instanceof WSAddressingFeature) {
+                        return (WSAddressingFeature)feature;
+                    }
+                }
+            }
+        }
+        return null;
+    }
+    /**
+     * If the isRequestor(message) == true and isAddressRequired() == false
+     * Assert all the wsa related assertion to true
+     * 
+     * @param message the current message
+     */
+    private void assertAddressing(Message message) {
+        AssertionInfoMap aim = message.get(AssertionInfoMap.class);
+        if (null == aim) {
+            return;
+        }
+        QName[] types = new QName[] {
+            MetadataConstants.ADDRESSING_ASSERTION_QNAME, MetadataConstants.USING_ADDRESSING_2004_QNAME,
+            MetadataConstants.USING_ADDRESSING_2005_QNAME, MetadataConstants.USING_ADDRESSING_2006_QNAME
+        };
+
+        for (QName type : types) {
+            assertAssertion(aim, type);
+            if (type.equals(MetadataConstants.ADDRESSING_ASSERTION_QNAME)) {
+                assertAssertion(aim, MetadataConstants.ANON_RESPONSES_ASSERTION_QNAME);
+                assertAssertion(aim, MetadataConstants.NON_ANON_RESPONSES_ASSERTION_QNAME);
+            } else if (type.equals(MetadataConstants.ADDRESSING_ASSERTION_QNAME_0705)) {
+                assertAssertion(aim, MetadataConstants.ANON_RESPONSES_ASSERTION_QNAME_0705);
+                assertAssertion(aim, MetadataConstants.NON_ANON_RESPONSES_ASSERTION_QNAME_0705);
+            }
+        }
+    }
+
     /**
      * Asserts all Addressing assertions for the current message, regardless their nested 
      * Policies.
@@ -417,10 +520,12 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                                  theMaps.getReplyTo(),
                                  theMaps.getFaultTo());
             }
-        } else if (!ContextUtils.isRequestor(message)) {            
+        } else if (!ContextUtils.isRequestor(message)) {
             //responder validates incoming MAPs
             AddressingPropertiesImpl maps = getMAPs(message, false, false);
+            //check responses          
             if (maps != null) {
+                checkAddressingResponses(maps.getReplyTo(), maps.getFaultTo());
                 assertAddressing(message, 
                                  maps.getReplyTo(),
                                  maps.getFaultTo());
@@ -444,10 +549,16 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                                                 message);
                 } 
                 if (!isOneway) {
-                    // ensure the inbound MAPs are available in both the full & fault
-                    // response messages (used to determine relatesTo etc.)
-                    ContextUtils.propogateReceivedMAPs(maps,
+                    // if ReplyTo address is none then 202 response status is expected
+                    // However returning a fault is more appropriate for request-response MEP
+                    if (ContextUtils.isNoneAddress(maps.getReplyTo())) {
+                        continueProcessing = false;
+                    } else {
+                        // ensure the inbound MAPs are available in both the full & fault
+                        // response messages (used to determine relatesTo etc.)
+                        ContextUtils.propogateReceivedMAPs(maps,
                                                        message.getExchange());
+                    }
                 }
             }
             if (continueProcessing) {
@@ -468,10 +579,74 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
             if (null != theMaps) {            
                 assertAddressing(message, theMaps.getReplyTo(), theMaps.getFaultTo());
             }
+            // If the wsa policy is enabled , but the client sets the
+            // WSAddressingFeature.isAddressingRequired to false , we need to assert all WSA assertion to true
+            if (!ContextUtils.isOutbound(message) && ContextUtils.isRequestor(message)
+                && getWSAddressingFeature(message) != null
+                && !getWSAddressingFeature(message).isAddressingRequired()) {
+                assertAddressing(message);
+            }
+            //CXF-3060 :If wsa policy is not enforced, AddressingProperties map is null and
+            // AddressingFeature.isRequired, requestor checks inbound message and throw exception
+            if (null == theMaps
+                && !ContextUtils.isOutbound(message) 
+                && ContextUtils.isRequestor(message)
+                && getWSAddressingFeature(message) != null
+                && getWSAddressingFeature(message).isAddressingRequired()) {
+                boolean missingWsaHeader = false;
+                AssertionInfoMap aim = message.get(AssertionInfoMap.class);
+                if (aim == null || aim.size() == 0) {
+                    missingWsaHeader = true;
+                }
+                if (aim != null && aim.size() > 0) {
+                    missingWsaHeader = true;
+                    QName[] types = new QName[] {
+                        MetadataConstants.ADDRESSING_ASSERTION_QNAME,
+                        MetadataConstants.USING_ADDRESSING_2004_QNAME,
+                        MetadataConstants.USING_ADDRESSING_2005_QNAME,
+                        MetadataConstants.USING_ADDRESSING_2006_QNAME
+                    };
+                    for (QName type : types) {
+                        for (AssertionInfo assertInfo : aim.getAssertionInfo(type)) {
+                            if (assertInfo.isAsserted()) {
+                                missingWsaHeader = false;
+                            }
+                        }
+                    }
+                }
+                if (missingWsaHeader) {
+                    String reason = BUNDLE.getString("MISSING_ACTION_MESSAGE");
+                    throw new SoapFault(reason, new QName(Names.WSA_NAMESPACE_NAME,
+                                                          Names.HEADER_REQUIRED_NAME));
+                }
+            }
         }
         return continueProcessing;
     }
 
+    private void checkAddressingResponses(EndpointReferenceType replyTo, EndpointReferenceType faultTo) {
+        if (this.addressingResponses.equals("ALL")) {
+            return;
+        }
+        boolean passed = false;
+        boolean anonReply = ContextUtils.isGenericAddress(replyTo);
+        boolean anonFault = ContextUtils.isGenericAddress(faultTo);
+        boolean isAnonymous = anonReply && anonFault;
+        if ("ANONYMOUS".equals(addressingResponses) && isAnonymous) {
+            passed = true;
+        } else if ("NON_ANONYMOUS".equals(addressingResponses)
+                   && (!anonReply && (faultTo.getAddress() != null && !anonFault) 
+                       || !anonReply && faultTo.getAddress() == null)) {
+            passed = true;
+        }
+        if (!passed) {
+            String reason = BUNDLE.getString("INVALID_ADDRESSING_PROPERTY_MESSAGE");
+            QName detail = "ANONYMOUS".equals(addressingResponses)
+                ? Names.ONLY_ANONYMOUS_ADDRESS_SUPPORTED_QNAME
+                : Names.ONLY_NONANONYMOUS_ADDRESS_SUPPORTED_QNAME;
+            throw new SoapFault(reason, detail);
+        }            
+    }
     /**
      * Perform MAP aggregation.
      *
@@ -524,7 +699,7 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
 
         if (inputMessage.getExtensionAttributes() != null) {
             String inputAction = ContextUtils.getAction(inputMessage);
-            if (inputAction != null) {
+            if (!StringUtils.isEmpty(inputAction)) {
                 return inputAction;
             }
         }
@@ -535,7 +710,7 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
         MessageInfo outputMessage = operation.getOutput();
         if (outputMessage != null && outputMessage.getExtensionAttributes() != null) {
             String outputAction = ContextUtils.getAction(outputMessage);
-            if (outputAction != null) {
+            if (!StringUtils.isEmpty(outputAction)) {
                 return outputAction;
             }
         }
@@ -562,7 +737,7 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                 if (isSameFault(faultInfo, faultName)) {
                     if (faultInfo.getExtensionAttributes() != null) {
                         String faultAction = ContextUtils.getAction(faultInfo);
-                        if (faultAction != null) {
+                        if (!StringUtils.isEmpty(faultAction)) {
                             return faultAction;
                         }
                     }
@@ -601,6 +776,23 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
         if (op.isUnwrapped()) {
             op = ((UnwrappedOperationInfo)op).getWrappedOperation();
         }
+        //CXF-2836:To correct the wsa:action header value for dispatch client
+        if (bop.getProperty("dispatchToOperation") != null) {
+            //modifies the bop and bp to the value of dispatch client really invokes, 
+            //This helps corrct the wsa:action header value
+            QName opName = (QName)bop.getProperty("dispatchToOperation");
+            OperationInfo dispatchOP = bop.getBinding().getService().getInterface().getOperation(opName);
+            BindingOperationInfo dispachBop = null;
+            for (BindingOperationInfo binfo : bop.getBinding().getOperations()) {
+                if (binfo.getOperationInfo().getName().toString().equals(opName.toString())) {
+                    dispachBop = binfo;
+                }
+            }
+            if (dispachBop != null && dispatchOP != null) {
+                bop = dispachBop;
+                op = dispatchOP;
+            }
+        }
         
         String actionUri = null;
         if (checkMessage) {
@@ -639,7 +831,7 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
             if (explicitAction != null) {
                 actionUri = explicitAction;
             } else if (null == op.getOutputName()) {
-                actionUri = addPath(opNamespace, op.getOutput().getName().getLocalPart());
+                actionUri = addPath(opNamespace, op.getName().getLocalPart() + "Response");
             } else {
                 actionUri = addPath(opNamespace, op.getOutputName());
             }
@@ -685,16 +877,28 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
             // add request-specific MAPs
             boolean isOneway = exchange.isOneWay();
             boolean isOutbound = ContextUtils.isOutbound(message);
-            Conduit conduit = null;
             
             // To
             if (maps.getTo() == null) {
+                Conduit conduit = null;
                 if (isOutbound) {
                     conduit = ContextUtils.getConduit(conduit, message);
                 }
+                String s = (String)message.get(Message.ENDPOINT_ADDRESS);
                 EndpointReferenceType reference = conduit != null
                                                   ? conduit.getTarget()
                                                   : ContextUtils.getNoneEndpointReference();
+                if (conduit != null && !StringUtils.isEmpty(s) 
+                    && !reference.getAddress().getValue().equals(s)) {
+                    EndpointReferenceType ref = new EndpointReferenceType();
+                    AttributedURIType tp = new AttributedURIType();
+                    tp.setValue(s);
+                    ref.setAddress(tp);
+                    ref.setMetadata(reference.getMetadata());
+                    ref.setReferenceParameters(reference.getReferenceParameters());
+                    ref.getOtherAttributes().putAll(reference.getOtherAttributes());
+                    reference = ref;
+                }
                 maps.setTo(reference);
             }
 
@@ -703,13 +907,7 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
             // current invocation
             EndpointReferenceType replyTo = maps.getReplyTo();
             if (ContextUtils.isGenericAddress(replyTo)) {
-                conduit = ContextUtils.getConduit(conduit, message);
-                if (conduit != null) {
-                    Destination backChannel = conduit.getBackChannel();
-                    if (backChannel != null) {
-                        replyTo = backChannel.getAddress();
-                    }
-                }
+                replyTo = getReplyTo(message, replyTo);
                 if (replyTo == null
                     || (isOneway
                         && (replyTo.getAddress() == null
@@ -749,6 +947,9 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                 && !Boolean.TRUE.equals(message.get(Message.PARTIAL_RESPONSE_MESSAGE))) {
                 String inMessageID = inMAPs.getMessageID().getValue();
                 maps.setRelatesTo(ContextUtils.getRelatesTo(inMessageID));
+            } else {
+                maps.setRelatesTo(ContextUtils
+                                  .getRelatesTo(Names.WSA_UNSPECIFIED_RELATIONSHIP));
             }
 
             // fallback fault action
@@ -766,6 +967,142 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
         }
     }
 
+    private EndpointReferenceType getReplyTo(Message message, 
+                                             EndpointReferenceType originalReplyTo) {
+        Exchange exchange = message.getExchange();
+        Endpoint info = exchange.get(Endpoint.class);
+        if (info == null) {
+            return originalReplyTo;
+        }
+        synchronized (info) {
+            EndpointInfo ei = info.getEndpointInfo();
+            Destination dest = ei.getProperty(DECOUPLED_DESTINATION, Destination.class);
+            if (dest == null) {
+                dest = createDecoupledDestination(message);
+                if (dest != null) {
+                    info.getEndpointInfo().setProperty(DECOUPLED_DESTINATION, dest);
+                }
+            }
+            if (dest != null) {
+                return dest.getAddress();
+            }
+        }
+        return originalReplyTo;
+    }
+
+    private Destination createDecoupledDestination(Message message) {
+        String replyToAddress = (String)message.getContextualProperty(WSAContextUtils.REPLYTO_PROPERTY);
+        if (replyToAddress != null) {
+            return setUpDecoupledDestination(message.getExchange().get(Bus.class),
+                                             replyToAddress, 
+                                             message);
+        }
+        return null;
+    }
+    /**
+     * Set up the decoupled Destination if necessary.
+     */
+    private Destination setUpDecoupledDestination(Bus bus, String replyToAddress, Message message) {        
+        EndpointReferenceType reference =
+            EndpointReferenceUtils.getEndpointReference(replyToAddress);
+        if (reference != null) {
+            String decoupledAddress = reference.getAddress().getValue();
+            LOG.info("creating decoupled endpoint: " + decoupledAddress);
+            try {
+                Destination dest = getDestination(bus, replyToAddress, message);
+                bus.getExtension(ClientLifeCycleManager.class).registerListener(DECOUPLED_DEST_CLEANER);
+                return dest;
+            } catch (Exception e) {
+                // REVISIT move message to localizable Messages.properties
+                LOG.log(Level.WARNING, 
+                        "decoupled endpoint creation failed: ", e);
+            }
+        }
+        return null;
+    }
+
+    /**
+     * @param address the address
+     * @return a Destination for the address
+     */
+    private Destination getDestination(Bus bus, String address, Message message) throws IOException {
+        Destination destination = null;
+        DestinationFactoryManager factoryManager =
+            bus.getExtension(DestinationFactoryManager.class);
+        DestinationFactory factory =
+            factoryManager.getDestinationFactoryForUri(address);
+        if (factory != null) {
+            Endpoint ep = message.getExchange().get(Endpoint.class);
+            
+            EndpointInfo ei = new EndpointInfo();
+            ei.setName(new QName(ep.getEndpointInfo().getName().getNamespaceURI(),
+                                 ep.getEndpointInfo().getName().getLocalPart() + ".decoupled"));
+            ei.setAddress(address);
+            destination = factory.getDestination(ei);
+            Conduit conduit = ContextUtils.getConduit(null, message);
+            if (conduit instanceof Observable) {
+                MessageObserver ob = ((Observable)conduit).getMessageObserver();
+                ob = new InterposedMessageObserver(bus, ob);
+                destination.setMessageObserver(ob);
+            }
+        }
+        return destination;
+    }
+    protected static class InterposedMessageObserver implements MessageObserver {
+        Bus bus;
+        MessageObserver observer;
+        public InterposedMessageObserver(Bus b, MessageObserver o) {
+            bus = b;
+            observer = o;
+        }
+        
+        /**
+         * Called for an incoming message.
+         * 
+         * @param inMessage
+         */
+        public void onMessage(Message inMessage) {
+            // disposable exchange, swapped with real Exchange on correlation
+            inMessage.setExchange(new ExchangeImpl());
+            inMessage.getExchange().put(Bus.class, bus);
+            inMessage.put(Message.DECOUPLED_CHANNEL_MESSAGE, Boolean.TRUE);
+            inMessage.put(Message.RESPONSE_CODE, HttpURLConnection.HTTP_OK);
+
+            // remove server-specific properties
+            //inMessage.remove(AbstractHTTPDestination.HTTP_REQUEST);
+            //inMessage.remove(AbstractHTTPDestination.HTTP_RESPONSE);
+            inMessage.remove(Message.ASYNC_POST_RESPONSE_DISPATCH);
+            updateResponseCode(inMessage);
+
+            //cache this inputstream since it's defer to use in case of async
+            try {
+                InputStream in = inMessage.getContent(InputStream.class);
+                if (in != null) {
+                    CachedOutputStream cos = new CachedOutputStream();
+                    IOUtils.copy(in, cos);
+                    inMessage.setContent(InputStream.class, cos.getInputStream());
+                }
+                observer.onMessage(inMessage);
+            } catch (IOException e) {
+                e.printStackTrace();
+            }
+        }
+        
+        private void updateResponseCode(Message message) {
+            Object o = message.get("HTTP.RESPONSE");
+            if (o != null) {
+                try {
+                    o.getClass().getMethod("setStatus", Integer.TYPE)
+                        .invoke(o, HttpURLConnection.HTTP_ACCEPTED);
+                } catch (Throwable t) {
+                    //ignore
+                }
+            }
+            
+        }
+        
+    }
+    
     /**
      * Get the starting point MAPs (either empty or those set explicitly
      * by the application on the binding provider request context).
@@ -818,13 +1155,9 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
         
         if (maps != null) {
             //WSAB spec, section 4.2 validation (SOAPAction must match action
-            Map<String, List<String>> headers 
-                = CastUtils.cast((Map<?, ?>)message.get(Message.PROTOCOL_HEADERS));
-            List<String> s = headers == null ? null : headers.get(Names.SOAP_ACTION_HEADER);
+            String sa = SoapActionInInterceptor.getSoapAction(message);
             String s1 = this.getActionUri(message, false);
-            if (s == null && headers != null) {
-                s = headers.get(Names.SOAP_ACTION_HEADER.toLowerCase());
-            }
+            
             if (maps.getAction() == null || maps.getAction().getValue() == null) {
                 String reason =
                     BUNDLE.getString("MISSING_ACTION_MESSAGE");
@@ -832,10 +1165,11 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                 ContextUtils.storeMAPFaultName(Names.HEADER_REQUIRED_NAME,
                                                message);
                 ContextUtils.storeMAPFaultReason(reason, message);
-                return false;
+                valid = false;
             }
-            if (s != null && s.size() > 0) {
-                String sa = s.get(0);
+            
+            if (!StringUtils.isEmpty(sa) && valid 
+                && !MessageUtils.isTrue(message.get(MAPAggregator.ACTION_VERIFIED))) {
                 if (sa.startsWith("\"")) {
                     sa = sa.substring(1, sa.lastIndexOf('"'));
                 }
@@ -844,12 +1178,12 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                     && !sa.equals(action)) {
                     //don't match, must send fault back....
                     String reason =
-                        BUNDLE.getString("INVALID_SOAPACTION_MESSAGE");
+                        BUNDLE.getString("INVALID_ADDRESSING_PROPERTY_MESSAGE");
     
                     ContextUtils.storeMAPFaultName(Names.ACTION_MISMATCH_NAME,
                                                    message);
                     ContextUtils.storeMAPFaultReason(reason, message);
-                    return false;
+                    valid = false;
                 } else if (!StringUtils.isEmpty(s1)
                     && !action.equals(s1)
                     && !action.equals(s1 + "Request")
@@ -866,20 +1200,37 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                     ContextUtils.storeMAPFaultName(Names.ACTION_NOT_SUPPORTED_NAME,
                                                    message);
                     ContextUtils.storeMAPFaultReason(reason, message);
-                    return false;
+                    valid = false;
                 }
-                
             }
             
+            AttributedURIType messageID = maps.getMessageID();
+            
+            if (!message.getExchange().isOneWay() 
+                    && (messageID == null || messageID.getValue() == null)
+                    && valid) {
+                String reason =
+                    BUNDLE.getString("MISSING_ACTION_MESSAGE");
+
+                ContextUtils.storeMAPFaultName(Names.HEADER_REQUIRED_NAME,
+                                               message);
+                ContextUtils.storeMAPFaultReason(reason, message);
+                
+                valid = false;
+            }
         
-            if (!allowDuplicates) {
-                AttributedURIType messageID = maps.getMessageID();
-                if (messageID != null
-                    && messageIDs.put(messageID.getValue(), 
-                                      messageID.getValue()) != null) {
-                    LOG.log(Level.WARNING,
-                            "DUPLICATE_MESSAGE_ID_MSG",
-                            messageID.getValue());
+            // Always cache message IDs, even when the message is not valid for some
+            // other reason.
+            if (!allowDuplicates && messageID != null && messageID.getValue() != null
+                && !messageIdCache.checkUniquenessAndCacheId(messageID.getValue())) {
+
+                LOG.log(Level.WARNING,
+                        "DUPLICATE_MESSAGE_ID_MSG",
+                        messageID.getValue());
+                
+                // Only throw the fault if something else has not already marked the
+                // message as invalid.
+                if (valid) {
                     String reason =
                         BUNDLE.getString("DUPLICATE_MESSAGE_ID_MSG");
                     String l7dReason = 
@@ -887,8 +1238,9 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
                     ContextUtils.storeMAPFaultName(Names.DUPLICATE_MESSAGE_ID_NAME,
                                                    message);
                     ContextUtils.storeMAPFaultReason(l7dReason, message);
-                    valid = false;
                 }
+                
+                valid = false;
             }
         } else if (usingAddressingAdvisory) {
             String reason =
@@ -897,8 +1249,13 @@ public class MAPAggregator extends AbstractPhaseInterceptor<Message> {
             ContextUtils.storeMAPFaultName(Names.HEADER_REQUIRED_NAME,
                                            message);
             ContextUtils.storeMAPFaultReason(reason, message);
-            return false;
+            valid = false;
         }
+        
+        if (Names.INVALID_CARDINALITY_NAME.equals(ContextUtils.retrieveMAPFaultName(message))) {
+            valid = false;
+        }
+        
         return valid;
     }
 }
